#! /usr/bin/env python3

"""Unit tests for Allgo module.

    Notes:
        cf https://docs.pytest.org/en/latest/contents.html
        tmp_path is a pytest fixture.
"""
# standard lib
import errno
import os
import sys
import tempfile

# not standard
import pytest

# local import
import allgo  # I need to modify constant
from allgo import *

if 2 == sys.version_info.major:
    from mock import Mock, patch
else:
    from unittest.mock import Mock, patch


# ================================================
class TestStatusError:

    REF_MSG = "whatever, an error message."

    @pytest.fixture
    def response(self):
        m = Mock()
        m.json.return_value = \
            { 'error': self.REF_MSG }
        return m

    def test_msg(self, response):
        err = StatusError(42, response)
        assert self.REF_MSG == err.msg


# ================================================
@pytest.fixture(scope="module")
def no_env_token():
    # setup
    try:
        os.environ.pop('ALLGO_TOKEN')  # , default=None) <= python3
    except KeyError:
        pass

    yield

    # teardown


# ================================================
def test_constants():
    assert MAIN_INSTANCE_URL.startswith("https://")
    assert TOKEN_FILE.endswith(".allgo_token")


# ------------------------------------------------
@pytest.mark.cstr
def test_cstr__no_token(no_env_token, tmp_path):
    allgo.TOKEN_FILE = os.path.join(str(tmp_path), 'fake')

    with pytest.raises(TokenError):
        Client()


@pytest.mark.cstr
def test_cstr__mini(no_env_token):
    assert isinstance(Client(token="jeton"), Client)


# ------------------------------------------------
@pytest.mark.getter
def test_token_getter__arg():
    token = "un_jeton"
    c = Client(token=token)
    assert token == c.token


@pytest.mark.getter
def test_token_getter__envvar():
    token = "un_jeton"
    os.environ['ALLGO_TOKEN'] = token
    c = Client()
    assert token == c.token


@pytest.mark.getter
def test_token_getter__file(no_env_token, tmp_path):
    token = "un_jeton"
    allgo.TOKEN_FILE = os.path.join(str(tmp_path), 'fake')
    with open(TOKEN_FILE, "w") as fs:
        fs.write(token)
    c = Client()
    assert token == c.token


# ------------------------------------------------
@pytest.fixture
def client():
    return Client(token="fake")


# ------------------------------------------------
@pytest.mark.getter
def test_allgo_url__default(client):
    assert MAIN_INSTANCE_URL == client.allgo_url


@pytest.mark.getter
def test_allgo_url__arg():
    token = "un_jeton"
    url = "a wanted url"
    c = Client(token=token, allgo_url=url)
    assert url == c.allgo_url


# ------------------------------------------------
@pytest.mark.getter
def test_verify_tls__default(client):
    assert True is client.verify_tls


@pytest.mark.getter
def test_verify_tls__arg():
    token  = "un_jeton"
    verify = False
    c = Client(token=token, verify_tls=verify)
    assert verify == c.verify_tls


# ------------------------------------------------
# -- Mock usage => https://realpython.com/testing-third-party-apis-with-mocks/
# -- json returned values are directly copied from allgo repository
# --    -> django/allgo/api/v1/views.py
# ------------------------------------------------
@pytest.mark.error
@patch('allgo.requests.post')
def test_create_job__user_not_exist(mock_post, client):
    # -- mock
    mock_post.return_value = Mock(status_code=401)
    # ~ mock_post.return_value.json.return_value = \
    # ~     {'error': 'API request without http authorisation'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.create_job('fake')

    err = err_info.value
    assert 401 == err.status_code
    # ~ assert '' == err.msg


@pytest.mark.error
@patch('allgo.requests.post')
def test_create_job__app_not_exist(mock_post, client):
    # -- mock
    mock_post.return_value = Mock(status_code=404)
    mock_post.return_value.json.return_value = \
        {'error': 'Application not found'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.create_job('fake')

    err = err_info.value
    assert 404 == err.status_code
    # ~ assert 'Application not found' == err.msg


@pytest.mark.error
@patch('allgo.requests.post')
def test_create_job__app_not_published(mock_post, client):
    # -- mock
    mock_post.return_value = Mock(status_code=404)
    mock_post.return_value.json.return_value = \
        {'error': "This app is not yet published"}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.create_job('fake')

    err = err_info.value
    assert 404 == err.status_code
    # ~ assert 'not yet published' in err.msg


@pytest.mark.error
@patch('allgo.requests.post')
def test_create_job__queue_not_exist(mock_post, client):
    # -- mock
    mock_post.return_value = Mock(status_code=400)
    mock_post.return_value.json.return_value = \
        {'error': 'Unknown queue'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.create_job('fake', queue='no_queue')

    err = err_info.value
    assert 400 == err.status_code
    # ~ assert 'Unknown queue' == err.msg


@pytest.mark.error
@patch('allgo.requests.post')
def test_create_job__param_error(mock_post, client):
    # -- mock
    mock_post.return_value = Mock(status_code=400)
    mock_post.return_value.json.return_value = \
        {'error': "Invalid parameters: blabla"}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.create_job('fake')

    err = err_info.value
    assert 400 == err.status_code
    # ~ assert err.msg.startswith('Invalid parameters')


@pytest.mark.success
@patch('allgo.Client._request_api')
def test_create_job__params_default(mock_request, client, tmp_path):
    """Here I'll check the params transmission."""
    app_name = 'fake'
    client.create_job(app_name)

    data = {  # default values, except for app_name
        "job[webapp_id]"  : app_name,
        "job[param]"      : '',
        "job[version]"    : None,
        "job[queue]"      : None,
    }
    file_dict = {}

    # -- tests
    mock_request.assert_called_once_with('post',
                                         url_end='jobs',
                                         data=data,
                                         files=file_dict)


@pytest.mark.success
@patch('allgo.Client._request_api')
def test_create_job__params(mock_request, client, tmp_path):
    """Here I'll check the params transmission.
        For that I'll mock the _request_api so that it returns
        the params = data and files dict."""
    # Mock
    def side_effect(method, url_end, **kwargs):
        return kwargs
    mock_request.side_effect = side_effect

    # create files to load
    filepaths = []
    for i, f in enumerate(['f1','f2','f3']):
        filepath = os.path.join(str(tmp_path), f )
        with open(filepath, "w") as fs:
            fs.write(f)  # content doesn't matter
        filepaths.append(filepath)

    app_name = 'fake'
    params   = 'un de kvar'
    version  = '1.42'
    queue    = 'quick'
    transmitted_params = client.create_job(app_name,
                                           version=version,
                                           params=params,
                                           files=filepaths,
                                           queue=queue)
    # expected
    data = {  # default values, except for app_name
        "job[webapp_id]"  : app_name,
        "job[param]"      : params,
        "job[version]"    : version,
        "job[queue]"      : queue,
    }
    file_dict_keys = [ "file[{}]".format(i) for i, _ in enumerate(filepaths) ]

    # -- tests
    assert 'data' in transmitted_params
    assert data == transmitted_params['data']
    assert 'files' in transmitted_params
    assert sorted(file_dict_keys) == sorted(transmitted_params['files'].keys())

    if 2 == sys.version_info.major:
        file_type = file
    else:
        from io import BufferedReader
        file_type = BufferedReader
    for handle in transmitted_params['files'].values():
        assert isinstance(handle, file_type)
        handle.close()


@pytest.mark.success
@patch('allgo.requests.post')
def test_create_job__response(mock_post, client):
    # -- mock
    job_id  = 33
    job_url = "http://job_url"
    post_output = {
        "avg_time": 0,  # legacy, not relevant anymore
        "id"      : job_id,
        "url"     : job_url,
    }
    mock_post.return_value = Mock(status_code=200)
    mock_post.return_value.json.return_value = post_output

    # -- tests
    response = client.create_job('fake')

    assert job_id  == response['id']
    assert job_url == response['url']


# ------------------------------------------------
@pytest.mark.error
@patch('allgo.requests.get')
def test_job_status__user_not_exist(mock_get, client):
    # -- mock
    mock_get.return_value = Mock(status_code=401)
    # ~ mock_get.return_value.json.return_value = \
    # ~     {'error': 'API request without http authorisation'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.job_status('fake_id')

    err = err_info.value
    assert 401 == err.status_code
    # ~ assert '' == err.msg


@pytest.mark.error
@patch('allgo.requests.get')
def test_job_status__unknown_id(mock_get, client):
    # -- mock
    mock_get.return_value = Mock(status_code=404)
    mock_get.return_value.json.return_value = \
        {'error': 'Job not found'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client.job_status('fake_id')

    err = err_info.value
    assert 404 == err.status_code
    # ~ assert 'Job not found' == err.msg


@pytest.mark.success
@patch('allgo.requests.get')
def test_job_status(mock_get, client):
    """The method return the request response as a dictionnary."""
    # -- mock
    job_id = 33
    files  = { 'f1': "file1_url", 'f2': "file2_url" }
    status = "bored"
    mock_get.return_value = Mock(status_code=200)
    mock_get.return_value.json.return_value = \
        {
            job_id: files,
            "status": status,
        }

    # -- tests
    response = client.job_status(job_id)

    assert files  == response[job_id]
    assert status == response['status']


# ------------------------------------------------
@pytest.mark.success
@patch('allgo.Client.create_job')
@patch('allgo.Client.job_status')
@patch('time.sleep')
def test_run_job__default(mock_sleep, mock_status, mock_create, client, capsys):
    """Here, I only want to test that both methods create_job and job_status are called.
        and job_status called several times.
        Each call to job_status will return the next element of the side_effect list.
        The Mock on sleep only intend to speed up the test."""
    # -- mock
    app = 'fake'
    fake_id = 6
    mock_create.return_value = { 'id': fake_id }
    job_status_returns = [
        { 'status': 'new' },
        { 'status': 'waiting' },
        { 'status': 'running' },
        { 'status': 'aborting' },
        { 'status': 'done' },
        { 'status': 'never_called' },  # should not be called
    ]
    mock_status.side_effect = job_status_returns

    output = client.run_job(app)

    # tests
    mock_create.assert_called_once()
    mock_status.assert_called_with(fake_id)

    x_calls = len(job_status_returns) - 1  # -1 because of 'never_called'
    assert x_calls == mock_status.call_count

    # run_job add the id to the job_status output
    x_output = { 'status': 'done', 'id': fake_id }
    assert x_output == output

    # without verbose, no output
    assert '' == capsys.readouterr().out


@pytest.mark.success
@patch('allgo.Client.create_job')
@patch('allgo.Client.job_status')
@patch('time.sleep')
def test_run_job__args(mock_sleep, mock_status, mock_create, client, capsys):
    """Here, I only want to test that both methods create_job and job_status are called.
        and job_status called several times.
        Each call to job_status will return the next element of the side_effect list."""
    # -- mock
    fake_id = 6
    mock_create.return_value = { 'id': fake_id }
    job_status_returns = [
        { 'status': 'new' },
        { 'status': 'waiting' },
        { 'status': 'waiting' },
        { 'status': 'waiting' },
        { 'status': 'done' },
        { 'status': 'never_called' },  # should not be called
    ]
    mock_status.side_effect = job_status_returns

    app = 'fake'
    kwargs = {
        'version': '666',
        'params' : '7',
        'files'  : ['truc', 'chose'],
    }
    sleep_d = 1
    client.run_job(app, sleep_duration=sleep_d, verbose=True, **kwargs)

    # tests
    mock_create.assert_called_once_with(app, **kwargs)
    mock_status.assert_called_with(fake_id)
    mock_sleep.assert_called_with(sleep_d)

    x_calls = len(job_status_returns) - 1  # -1 because of 'never_called'
    assert x_calls == mock_status.call_count

    x_output = "\nnew\t\nwaiting\t..\ndone\t\n"
    assert x_output == capsys.readouterr().out


# ------------------------------------------------
@pytest.mark.error
@patch('allgo.requests.get')
def test__request_file(mock_get, client):
    """assert a StatusError is raised if the request fail.
        The error message is not checked."""
    # -- mock
    err_code = 666
    mock_get.return_value = Mock(status_code=err_code)
    # ~ mock_get.return_value.json.return_value = \
    # ~     {'error': 'API request without http authorisation'}

    # -- tests
    with pytest.raises(StatusError) as err_info:
        client._request_file('url_bidon')

    err = err_info.value
    assert err_code == err.status_code
    # ~ assert '' == err.msg


@pytest.mark.error
def test_app_metrics(client):
    """app_metrics method is not yet implemented."""
    with pytest.raises(NotImplementedError):
        client.app_metrics('fake', 'what')


# ------------------------------------------------
def __mock_file(mock_get):
    file_content = "fake line1.\nline2 is fake too.\n"

    # -- mock
    mock_get.return_value = Mock(status_code=200)
    mock_get.return_value.iter_content.return_value = \
        iter([file_content.encode()])

    return file_content


@pytest.mark.success
@patch('allgo.requests.get')
def test_download_file__default(mock_get, client):
    """The method returns the filepath where the file has been written,
        outdir is the localdir by default."""
    # -- mock
    file_content = __mock_file(mock_get)

    # -- tests
    filename = 'wanted_file'
    file_url = "https://whatever.fr/" + filename
    filepath = client.download_file(file_url)

    assert "./" + filename == filepath

    content = ""
    with open(filepath, "r") as file_out:
        content = file_out.read()
    assert file_content == content

    # cleaning
    os.remove(filepath)


@pytest.mark.success
@patch('allgo.requests.get')
def test_download_file__outdir(mock_get, client, tmp_path):
    """The method returns the filepath where the file has been written."""
    # -- mock
    file_content = __mock_file(mock_get)

    # -- tests
    filename = "wanted_file"
    file_url = "https://whatever.fr/" + filename
    filepath = client.download_file(file_url, outdir=tmp_path)

    assert os.path.join(str(tmp_path), filename) == filepath

    content = ""
    with open(filepath, "r") as file_out:
        content = file_out.read()
    assert file_content == content


@pytest.mark.error
@patch('allgo.requests.get')
def test_download_file__outdir_already_exists(mock_get, client, tmp_path):
    """The method should not raise an error if the output dir already exist."""
    # -- mock
    __mock_file(mock_get)

    # -- tests
    file_url = "https://whatever.fr/x"
    try:
        client.download_file(file_url, outdir=tmp_path)
    except Exception:
        pytest.fail("In download_file, existing outdir should not raise an error.")


@pytest.mark.error
@patch('allgo.os.makedirs')
@patch('allgo.requests.get')
def test_download_file__outdir_error(mock_get, mock_mkdir, client, tmp_path):
    """The method should raise an error if the output dir cannot be created."""
    # -- mock
    __mock_file(mock_get)

    err_code = errno.EPERM
    err_msg  = 'fake msg'
    mock_mkdir.side_effect = OSError(err_code, err_msg)

    # -- tests
    file_url = "https://whatever.fr/x"
    with pytest.raises(OSError) as err_info:
        client.download_file(file_url, outdir=tmp_path)

    err = err_info.value
    assert err_code == err.errno
    assert err_msg  == err.strerror


@pytest.mark.success
@patch('allgo.requests.get')
def test_download_file__force(mock_get, client):
    """If the output file already exists,
        the force param, False by default result in an OSError,
        if force is set to True, no error should be raised.
    """
    # -- mock
    file_content = __mock_file(mock_get)

    # -- tests
    # create the file
    tmp_filepath = tempfile.NamedTemporaryFile(dir='.')
    filename = os.path.basename(tmp_filepath.name)
    file_url = "https://whatever.fr/" + filename

    # default is to raise an exception if the file exists.
    # ~ with pytest.raises(FileExistsError): # python3
    with pytest.raises(OSError) as err_info:
        filepath = client.download_file(file_url, force=False)  # default

    err = err_info.value
    assert errno.EEXIST == err.errno

    try:
        client.download_file(file_url, force=True)
    # ~ except FileExistsError:  #python3
    except OSError as e:
        tmp_filepath.close()
        if e.errno == errno.EEXIST:
            pytest.fail("download_file call with force=True failed.")

    # cleaning
    tmp_filepath.close()


# ------------------------------------------------
@pytest.mark.error
@patch('allgo.Client.job_status')
def test_get_log__job_not_done_yet(mock_status, client):
    fake_id = '6'
    mock_status.return_value = \
        {
            fake_id: { 'allgo.log': 'http://fake.com/allgo.log' },
            "status": 'running',
        }

    # test
    with pytest.raises(AllgoError):
        # Note: calling get_log return a generator
        #   => the code is not executed yet !
        #   => the generator has to be used to execute the code
        list(client.get_log(fake_id))


@pytest.mark.error
@patch('allgo.requests.get')
def test_get_log__download_error(mock_get, client):
    """I'll not test all error status, only error type."""
    fake_id = '6'
    mock_get.return_value = Mock(status_code=404)

    # test
    with pytest.raises(StatusError):
        # Note: calling get_log return a generator
        #   => the code is not executed yet !
        #   => the generator has to be used to execute the code
        list(client.get_log(fake_id))


@pytest.mark.success
@patch('allgo.Client.job_status')
@patch('allgo.requests.get')
def test_get_log__small(mock_get, mock_status, client):
    """I reused the mock_file used for test_download_file.
        The whole content can be retrieve in one get."""
    # -- mock
    fake_id = '6'
    mock_status.return_value = \
        {
            fake_id: { 'allgo.log': 'http://fake.com/allgo.log' },
            "status": 'done',
        }
    file_content = __mock_file(mock_get)

    # -- tests
    # Note: calling get_log return a generator
    #   => the code is not executed yet !
    #   => the generator has to be used to execute the code
    import types
    g = client.get_log(fake_id)
    assert isinstance(g, types.GeneratorType)
    for content in g:
        assert file_content == content


@pytest.mark.success
@patch('allgo.Client.job_status')
@patch('allgo.requests.get')
def test_get_log__big(mock_get, mock_status, client):
    """Here, i test that the i."""
    # -- mock
    fake_id = '6'
    mock_status.return_value = \
        {
            fake_id: { 'allgo.log': 'http://fake.com/allgo.log' },
            "status": 'done',
        }

    file_content = []
    alphabet = 'abcde'
    for c in alphabet:
        chunk = ''
        for i in range(8):
            chunk += c * 10 + "\n"
        file_content.append(chunk)
    # file_content is 5 chunks of 8 lines each
    #   and each line is the same letter repeated 10 times.

    strings = [s.encode() for s in file_content]
    mock_get.return_value = Mock(status_code=200)
    mock_get.return_value.iter_content.return_value = \
        iter(strings)

    # -- tests
    # Note: calling get_log return a generator
    #   => the code is not executed yet !
    #   => the generator has to be used to execute the code
    for i, content in enumerate(client.get_log(fake_id)):
        assert file_content[i] == content


# ------------------------------------------------
@pytest.mark.success
@patch('allgo.requests.get')
def test__request_api__manage_ValueErrors(mock_get, client):
    """_request_api method should manage well error that can occurr in querying.
        if ValueError the response should be returned.

        ValueError happens when response content cannot be converted to json."""
    # -- mock
    response = mock_get.return_value
    response.status_code = 200
    response.json = Mock(side_effect=ValueError())

    # -- tests
    resp = client._request_api('get', 'fin_url')
    assert response == resp


@pytest.mark.error
def test__request_api__manage_KeyErrors(client):
    """_request_api method should manage well error that can occurr in querying.
        if KeyError is raised, an AllgoError should be raised.

        KeyError happens when an unknown actions is used as param."""

    # -- tests
    with pytest.raises(AllgoError):
        client._request_api('fail', 'fin_url')
